#! /usr/bin/env python
# -*- coding: utf-8 -*-

import json
from collections import defaultdict
# part of module
from .concordances import df_node_to_concordance


# corrections
def apply_correction(row, correction):
    value, lower_bound, upper_bound = row
    value += correction
    if value < lower_bound or value > upper_bound:
        value = -1
    return value


def apply_corrections(df_anchor, corrections):
    for correction in corrections:
        if correction[0] in df_anchor.columns:
            df_anchor[correction[0]] = df_anchor[
                [correction[0], 'region_start', 'region_end']
            ].apply(lambda x: apply_correction(x, correction[1]), axis=1)
    return df_anchor


# hole structure
def get_holes(df, anchor_holes, region_holes):

    holes = defaultdict(dict)

    for idx in anchor_holes.keys():
        anchor = anchor_holes[idx]
        row_nr = int(df[df['anchor'] == anchor].index.values)
        word = df.at[row_nr, 'word']
        if 'lemma' in df.columns:
            lemma = str(df.at[row_nr, 'lemma'])
        else:
            lemma = None

        holes['words'][idx] = word
        holes['lemmas'][idx] = lemma

    for idx in region_holes.keys():
        region_start = region_holes[idx][0]
        region_end = region_holes[idx][1]

        row_start = df[df['anchor'] == region_start]
        row_end = df[df['anchor'] == region_end]

        # if both of them are empty
        if row_start.empty and row_end.empty:
            words = None
            lemmas = None
        else:
            # if one of them is empty: start and end are the same
            if row_start.empty:
                row_start = row_end
            elif row_end.empty:
                row_end = row_start

            row_start = int(row_start.index.values)
            row_end = int(row_end.index.values)
            region = df.loc[row_start:row_end]
            words = " ".join(list(region['word']))
            if 'lemma' in df.columns:
                lemmas = " ".join(list(region['lemma']))
            else:
                lemmas = None

        holes['words'][idx] = words
        holes['lemmas'][idx] = lemmas

    return holes


# actual query
def anchor_query(engine, query_str, anchors, regions, s_break,
                 context=None, p_show=['lemma'],
                 match_strategy="longest"):

    # get hole structure
    anchor_holes = dict()
    for anchor in anchors:
        if anchor[2] is not None:
            anchor_holes[anchor[2]] = anchor[0]
    region_holes = dict()
    for region in regions:
        if region[2] is not None:
            region_holes[region[2]] = region[:2]

    # get df_anchor
    df_anchor = engine.df_anchor_from_query(
        query_str,
        s_break=s_break,
        context=context,
        match_strategy=match_strategy
    )

    # initialize output
    result = dict()
    result['matches'] = list()
    result['holes'] = defaultdict(list)

    # if result is empty ...
    if df_anchor.empty:
        result['nr_matches'] = 0
        return result

    # apply corrections
    df_anchor = apply_corrections(df_anchor, anchors)

    # retrieve concordance
    concordance = df_node_to_concordance(engine,
                                         df_anchor,
                                         context=100,
                                         order="first",
                                         cut_off=None,
                                         p_show=p_show,
                                         anchored=True)

    # number of matches
    result['nr_matches'] = len(concordance)

    # loop through concordances
    for key in concordance.keys():

        line = concordance[key]

        # fill concordance line
        entry = dict()
        entry['df'] = line.to_dict()
        entry['position'] = key
        entry['full'] = " ".join(entry['df']['word'].values())

        # hole structure
        holes = get_holes(line, anchor_holes, region_holes)
        if 'lemmas' in holes.keys():
            entry['holes'] = holes['lemmas']
        else:
            entry['holes'] = holes['words']

        result['matches'].append(entry)

        # append to global holes list
        for idx in entry['holes'].keys():
            result['holes'][idx].append(entry['holes'][idx])

    return result


# read file and process query
def process_argmin_file(engine, query_path, concordance_settings):

    with open(query_path, "rt") as f:
        try:
            query = json.loads(f.read())
        except json.JSONDecodeError:
            print("WARNING: not a valid json file")
            return

    # add query parameters
    query['corpus_name'] = engine.corpus_name
    query['concordance_settings'] = concordance_settings
    query['query_path'] = query_path

    # get result
    query['result'] = anchor_query(engine,
                                   query['query'],
                                   query['anchors'],
                                   query['regions'],
                                   concordance_settings['s_break'],
                                   concordance_settings['context'],
                                   concordance_settings['p_show'],
                                   concordance_settings['match_strategy'])

    return query



#! /usr/bin/env python
# -*- coding: utf-8 -*-

from random import sample
# part of module
from .cwb import anchor_query_to_anchors
# requirements
from pandas import DataFrame


class Concordance(object):

    def __init__(self,
                 engine,
                 context=20,
                 s_break='text',
                 order='random',
                 p_show=[],
                 cut_off=100):

        self.engine = engine

        self.settings = {
            'context': context,
            's_break': s_break,
            'p_show': p_show,
            'order': order,
            'cut_off': cut_off
        }

    def query(self, query, anchored=None):

        if anchored is None:
            anchored = len(anchor_query_to_anchors(query)) > 0

        if not anchored:
            df_node = self.engine.df_node_from_query(
                query,
                self.settings['s_break']
            )
        else:
            df_node = self.engine.df_anchor_from_query(
                query,
                self.settings['s_break']
            )

        if len(df_node) == 0:
            return {}

        concordance = df_node_to_concordance(
            engine=self.engine,
            df_node=df_node,
            context=self.settings['context'],
            order=self.settings['order'],
            cut_off=self.settings['cut_off'],
            p_show=self.settings['p_show'],
            anchored=anchored
        )

        return concordance


# node to concordance (needs engine) ###########################################
def df_node_to_concordance(engine,
                           df_node,
                           context=20,
                           order='random',
                           cut_off=100,
                           p_show=[],
                           anchored=None,
                           role='match'):
    """ formats concordance lines for a df_node """

    # avoid trying to get more concordance lines than there are
    topic_matches = set(df_node['match_start'])
    if not cut_off or len(topic_matches) < cut_off:
        cut_off = len(topic_matches)

    # take appropriate sub-set
    if order == 'random':
        topic_matches_cut = sample(topic_matches, cut_off)
    elif order == 'first':
        topic_matches_cut = sorted(list(topic_matches))[:cut_off]
    elif order == 'last':
        topic_matches_cut = sorted(list(topic_matches))[-cut_off:]
    else:
        raise NotImplementedError("concordance order not implemented")

    # check if there's anchors
    anchor_keys = set(df_node.columns) - {'region_start', 'region_end', 's_id',
                                          'match_end', 'match_start'}

    # automatically determine by default
    if anchored is None:
        anchored = len(anchor_keys) > 0

    # fill concordance dictionary
    concordance = dict()
    for match_start in topic_matches_cut:

        # take values from row
        row = df_node.loc[df_node['match_start'] == match_start]
        # int(row.pop('match_start'))
        match_end = int(row.pop('match_end'))
        # s_id = row.pop('s_id')
        s_start = int(row.pop('region_start'))
        s_end = int(row.pop('region_end'))

        # handle optional anchors
        if anchored:
            anchors = dict()
            for anchor in anchor_keys:
                anchors[anchor] = int(row[anchor])

        # init concordance line variables
        roles = list()
        offset = list()
        cpos = list()

        # fill cpos, role, offset
        for position in range(
                max([s_start, match_start-context]),
                min([match_end+context, s_end])+1
        ):
            cpos.append(position)
            roles.append(position >= match_start and position <= match_end)
            if position < match_start:
                offset.append(position - match_start)
            elif position > match_end:
                offset.append(position - match_end)
            else:
                offset.append(0)

        # lexicalize positions
        word = engine.lexicalize_positions(cpos)

        # transform concordance line to dataframe
        concordance_line = DataFrame(
            index=cpos,
            data={
                'word': word,
                role: roles,
                'offset': offset
            }
        )
        concordance_line.index.name = 'cpos'

        # handle optional anchors
        if anchored:
            concordance_line['anchor'] = None
            for anchor in anchors.keys():
                if anchors[anchor] != -1:
                    concordance_line.at[anchors[anchor], 'anchor'] = anchor

        # handle additional p-attributes
        for p_att in p_show:
            p_show_tokens = engine.lexicalize_positions(cpos, p_att)
            concordance_line[p_att] = p_show_tokens

        # save concordance line
        concordance[match_start] = concordance_line

    return concordance



#! /usr/bin/env python
# -*- coding: utf-8 -*-

# Name of this module: PyCQP
# Version 2.0 (Febr. 2008)
# Joerg Asmussen, DSL
# Some changes by Philipp Heinrich (2019)


import sys
import os
import re
import random
import time
import subprocess
import select
from six.moves import _thread as thread


# GLOBAL CONSTANTS OF MODULE:
cProgressControlCycle = 30  # secs between each progress control cycle
cMaxRequestProcTime = 300   # max secs for processing a user request


# ERROR MESSAGE TYPES:
class ErrCQP:
    """Handle CQP error message."""
    def __init__(self, msg):
        """Class constructor."""
        self.msg = msg.rstrip()


class ErrKilled:
    """Handle errors when process is killed."""
    def __init__(self, msg):
        """Class constructor."""
        self.msg = msg.rstrip()


# ACTUAL INTERFACE:
class CQP:
    """Wrapper for CQP."""

    def _progressController(self):
        """Control the progess.

        CREATED: 2008-02

        This method is run as a thread.
        At certain intervals (cProgressControlCycle), it controls how long the
        CQP process of the current user has spent on processing this user's
        latest CQP command. If this time exceeds a certain maximun
        (cMaxRequestProcTime), this method kills the CQP process.
        """
        self.runController = True
        while self.runController:
            time.sleep(cProgressControlCycle)
            if self.execStart is not None:
                if time.time() - self.execStart > cMaxRequestProcTime *\
                 self.maxProcCycles:
                    print(
                        '''WARNING!: PROGRESS CONTROLLER IDENTIFIED BLOCKING CQP
                        PROCESS ID {}'''.format(self.CQP_process.pid), end='')
                    # os.kill(self.CQP_process.pid, SIGKILL) - doesn't work!
                    os.popen("kill -9 " + str(self.CQP_process.pid))  # works!
                    print("=> KILLED!")
                    self.CQPrunning = False
                    break

    def __init__(self, bin="/usr/local/bin/cqp", options='-c'):
        """Class constructor."""
        self.execStart = time.time()
        self.maxProcCycles = 1.0
        # start CQP as a child process of this wrapper
        if bin is None:
            print("ERROR: Path to CQP binaries undefined", file=sys.stderr)
            sys.exit(1)
        self.CQP_process = subprocess.Popen(bin + ' ' + options,
                                            shell=True,
                                            stdin=subprocess.PIPE,
                                            stdout=subprocess.PIPE,
                                            stderr=subprocess.PIPE,
                                            universal_newlines=True,
                                            close_fds=True)
        self.CQPrunning = True
        thread.start_new_thread(self._progressController, ())
        # "cqp -c" should print version on startup:
        version_string = self.CQP_process.stdout.readline()
        version_string = version_string.rstrip()  # Equivalent to Perl's chomp
        self.CQP_process.stdout.flush()
        print(version_string, file=sys.stderr)
        version_regexp = re.compile(r'^CQP\s+(?:\w+\s+)*([0-9]+)\.([0-9]+)(?:\.b?([0-9]+))?(?:\s+(.*))?$')
        match = version_regexp.match(version_string)
        if not match:
            print("ERROR: CQP backend startup failed", file=sys.stderr)
            sys.exit(1)
        self.major_version = int(match.group(1))
        self.minor_version = int(match.group(2))
        self.beta_version = int(match.group(3))
        self.compile_date = match.group(4)

        # We need cqp-2.2.b41 or newer (for query lock):
        if not (
                self.major_version >= 3 or
                (self.major_version == 2 and
                    self.minor_version == 2 and
                    self.beta_version >= 41)
                ):
            print(
                "ERROR: CQP version too old: " + version_string,
                file=sys.stderr)
            sys.exit(1)

        # Error handling:
        self.error_handler = None
        self.status = 'ok'
        self.error_message = ''  # we store compound error messages as a STRING
        self.errpipe = self.CQP_process.stderr.fileno()

        # Debugging (prints more or less everything on stdout)
        self.debug = False

        # CQP defaults:
        self.Exec('set PrettyPrint off')
        self.execStart = None

    def Terminate(self):
        """Terminate controller thread, must be called before deleting CQP."""
        self.execStart = None
        self.runController = False

    def SetProcCycles(self, procCycles):
        """Set procCycles."""
        print("    Setting procCycles to {}".format(procCycles))
        self.maxProcCycles = procCycles
        return int(self.maxProcCycles * cMaxRequestProcTime)

    def __del__(self):
        """Stop running CQP instance."""
        if self.CQPrunning:
            # print "Deleting CQP with pid", self.CQP_process.pid, "...",
            self.CQPrunning = False
            self.execStart = time.time()
            if self.debug:
                print("Shutting down CQP backend ...", end='')
            self.CQP_process.stdin.write('exit;')  # exits CQP backend
            if self.debug:
                print("Done\nCQP object deleted.")
            self.execStart = None
            # print "Finished"

    def Exec(self, cmd):
        """Execute CQP command.

        The method takes as input a command string and sends it
        to the CQP child process
        """
        self.execStart = time.time()
        self.status = 'ok'
        cmd = cmd.rstrip()  # Equivalent to Perl's 'chomp'
        cmd = re.sub(r';\s*$', r'', cmd)
        if self.debug:
            print("CQP <<", cmd + ";")
        try:
            self.CQP_process.stdin.write(cmd + '; .EOL.;\n')
        except IOError:
            return None
        # In CQP.pm lines are appended to a list @result.
        # This implementation prefers a string structure instead
        # because output from this module is meant to be transferred
        # accross a server connection. To enable the development of
        # client modules written in any language, the server only emits
        # strings which then are to be structured by the client module.
        # The server does not emit pickled data according to some
        # language dependent protocol.
        self.CQP_process.stdin.flush()
        result = []
        while self.CQPrunning:
            ln = self.CQP_process.stdout.readline()
            ln = ln.strip()  # strip off whitespace from start and end of line
            if re.match(r'-::-EOL-::-', ln):
                if self.debug:
                    print("CQP " + "-" * 60)
                break
            if self.debug:
                print("CQP >> " + ln)
            if ln != '':
                result.append(ln)
            self.CQP_process.stdout.flush()
        self.Checkerr()
        self.execStart = None
        result = '\n'.join(result)
        result = result.rstrip()  # strip off whitespace from EOL (\n)
        return result

    def Query(self, query):
        """Execute query in safe mode (query lock)."""
        result = []
        key = str(random.randint(1, 1000000))
        errormsg = ''  # collect CQP error messages AS STRING
        ok = True      # check if any error occurs
        self.Exec('set QueryLock ' + key)  # enter query lock mode
        if self.status != 'ok':
            errormsg = errormsg + self.error_message
            ok = False
        result = self.Exec(query)
        if self.status != 'ok':
            errormsg = errormsg + self.error_message.decode('utf-8')
            ok = ok and False
        self.Exec('unlock ' + key)  # unlock with random key
        if self.status != 'ok':
            errormsg = errormsg + self.error_message
            ok = ok and False
        # Set error status & error message:
        if ok:
            self.status = 'ok'
        else:
            self.status = 'error'
            self.error_message = errormsg
        return result

    def Dump(self, subcorpus='Last', first=None, last=None):
        """Dump named query result into table of corpus positions."""
        if first is None and last is None:
            result = self.Exec('dump ' + subcorpus + ";")
        elif ((not isinstance(first, int) and first is not None) or
                (not isinstance(last, int) and last is not None)):
            sys.stderr.write(
                            "ERROR: Invalid value for first (" +
                            str(first) + ") or last (" + str(last) +
                            ") line in Dump() method\n")
            sys.exit(1)
        elif isinstance(first, int) and isinstance(last, int):
            if first > last:
                sys.stderr.write(
                    "ERROR: Invalid value for first line (first = " +
                    str(first) + " > last = " + str(last) +
                    ") in Dump() method\n")
                sys.exit(1)
            else:
                result = self.Exec(
                    'dump ' +
                    subcorpus +
                    " " +
                    str(first) +
                    " " +
                    str(last))
        else:
            if first is not None and last is None:
                last = first
            elif last is not None and first is None:
                first = last
            result = self.Exec(
                'dump ' +
                subcorpus +
                " " +
                str(first) +
                " " +
                str(last) +
                ";")
        result = [x.split('\t') for x in result.split('\n')]
        return result

    def Undump(self, subcorpus='tmp', table=[], tmp_path="/tmp/cwb_undump"):
        """Undump named query result from table of corpus positions."""
        wth = ''  # undump with target and keyword
        n_el = None  # number of anchors for each match (from first row)
        n_matches = len(table)  # number of matches (= remaining arguments)
        # We have to read undump table from temporary file:
        with open(tmp_path, "wt") as tf:
            tf.write(str(n_matches) + '\n')
            for row in table:
                row_el = len(row)
                if n_el is None:
                    n_el = row_el
                    if (n_el < 2) or (n_el > 4):
                        raise ValueError(
                            "ERROR: Row arrays in undump table must have " +
                            "between 2 and 4 elements (first row has " +
                            str(n_el) + " elements)"
                        )
                    if n_el >= 3:
                        wth = 'with target'
                    if n_el == 4:
                        wth = wth + ' keyword'
                elif row_el != n_el:
                    raise ValueError(
                        "ERROR: All rows in undump table must have the same " +
                        "length (first row = " + str(n_el) + ", this row = " +
                        str(row_el) + ")"
                    )
                tf.write('\t'.join(row) + '\n')
        # Send undump command with filename of temporary file:
        self.Exec("undump " + subcorpus + " " + wth + " < '" + tmp_path + "';")

    def Group(self, subcorpus='Last',
              spec1='match.word', spec2='', cutoff='1'):
        """Compute frequency distribution over attribute values.

        (single values or pairs) using group command.

        Note that the arguments are specified in the logical order,
        in contrast to "group"
        """
        spec2_regexp = re.compile(r'^[0-9]+$')
        if spec2_regexp.match(spec2):
            cutoff = spec2
            spec2 = ''
        spec_regexp = re.compile(
          r'^(match|matchend|target[0-9]?|keyword)\.([A-Za-z0-9_-]+)$')
        match = re.match(spec_regexp, spec1)
        if not match:
            print(
                "ERROR: Invalid key '" + spec1 + "' in Group() method",
                file=sys.stderr)
            sys.exit(1)
        spec1 = match.group(1) + ' ' + match.group(2)
        if spec2 != '':
            match = re.match(spec_regexp, spec2)
            if not match:
                print(
                    "ERROR: Invalid key '" + spec2 + "' in Group() method",
                    file=sys.stderr)
                sys.exit(1)
            spec2 = match.group(1) + ' ' + match.group(2)
            cmd = 'group ' + subcorpus + ' ' + spec2 + ' by ' + spec1 + \
                  ' cut ' + cutoff
        else:
            cmd = 'group ' + subcorpus + ' ' + spec1 + ' cut ' + cutoff
        result = self.Exec(cmd)
        return result

    def Count(self, subcorpus='Last', sort_clause=None, cutoff=1):
        """Compute frequency distribution for match strings.

        Based on sort clause.
        """
        if sort_clause is None:
            print(
                "ERROR: Parameter 'sort_clause' undefined in Count() method",
                file=sys.stderr)
            sys.exit(1)
        return self.Exec(
            'count ' +
            subcorpus +
            ' by ' +
            sort_clause +
            ' cut ' +
            str(cutoff))

    def Checkerr(self):
        """Check CQP's stderr stream for error messages.

        (returns true if there was an error).
        OBS! In CQP.pm the error_message is stored in a list,
        In PyCQP_interface.pm we use a string (which better can be sent
        accross the server line).
        """
        ready = select.select([self.errpipe], [], [], 0)
        if self.errpipe in ready[0]:
            # We've got something on stderr -> an error must have occurred:
            self.status = 'error'
            self.error_message = self.Readerr()
        return not self.Ok()

    def Readerr(self):
        """Read all available lines from CQP's stderr stream."""
        return os.read(self.errpipe, 16384)

    def Status(self):
        """Read the CQP object's (error) status."""
        return self.status

    def Ok(self):
        """Simplified interface for checking for CQP errors."""
        if self.CQPrunning:
            return (self.Status() == 'ok')
        else:
            return False

    def Error_message(self):
        """Return the CQP error message."""
        if self.CQPrunning:
            return ErrCQP(self.error_message)
        else:
            msgKilled = '**** CQP KILLED ***\n\
            CQP COULD NOT PROCESS YOUR REQUEST\n'
            return ErrKilled(msgKilled + self.error_message)

    def Error(self, msg):
        """Processe/output error messages.

        (optionally run through user-defined error handler)
        """
        if self.error_handler is not None:
            self.error_handler(msg)
        else:
            print(msg, file=sys.stderr)

    def Set_error_handler(self, handler=None):
        """Set user-defined error handler."""
        self.error_handler = handler

    def Debug(self, on=False):
        """Switch debugging output on/off."""
        prev = self.debug
        self.debug = on
        return prev



#! /usr/bin/env python
# -*- coding: utf-8 -*-

import re
import os
from glob import glob
from collections import Counter
# part of module
from .cqp_interface import CQP
from .utils import Cache, create_identifier
# requirements
from pandas import DataFrame, read_csv, to_numeric
from CWB.CL import Corpus
from io import StringIO


# some utilities that don't need the engine
def anchor_query_to_anchors(anchor_query, strict=False):
    """ get anchors present in query """

    anchors = dict()
    p = re.compile(r"@\d")

    for m in p.finditer(anchor_query):
        if strict:
            if m.group() in anchors.keys():
                raise KeyError('duplicate keys provided')
        anchors[m.group()] = [m.start(), m.end()]

    keys = [int(a[1]) for a in anchors.keys()]

    return keys


def dump_to_df_node(dump, nr_columns=4):
    """ convert cqp output to pandas dataframe """
    if nr_columns == 4:
        df = read_csv(StringIO(dump),
                      sep="\t", header=None, index_col=[0, 1],
                      names=["match_start", "match_end", "target", "keyword"],
                      dtype=int)
    elif nr_columns == 2:
        df = read_csv(StringIO(dump),
                      sep="\t", header=None,
                      names=["target", "keyword"],
                      dtype=int)
    return df


class CWBEngine(object):
    """ interface to CWB, convenience wrapper """

    def __init__(self,
                 corpus_name,
                 registry_path="/usr/local/share/cwb/registry/",
                 lib_path=None,
                 cqp_bin="cqp",
                 cache_path="/tmp/ccc-cache"):
        """Establishes connection to indexed corpus. Raises KeyError if corpus
        not in registry.
        """

        # set cache
        self.cache = Cache("/tmp/ccc-cache")

        # registry path
        self.registry_path = registry_path

        # obligatory corpus settings
        self.corpus_name = corpus_name

        # set interface
        self.corpus = Corpus(
            self.corpus_name,
            registry_dir=self.registry_path
        )
        self.cqp = CQP(
            bin=cqp_bin,
            options='-c -r ' + self.registry_path
        )

        # activate corpus
        self.cqp.Exec(self.corpus_name)

        # get corpus description
        df = read_csv(StringIO(self.cqp.Exec("show cd;")), sep="\t",
                      names=['att', 'value', 'annotation'])
        self.corpus_p_att = list(df[df['att'] == "p-Att"]['value'])
        s_att = df[df['att'] == "s-Att"]
        self.corpus_s_att_with = list(s_att[s_att['annotation'] == "-V"]['value'])
        self.corpus_s_att_without = list(s_att[s_att['annotation'] != "-V"]['value'])

        # get corpus info
        self._N = len(self.corpus.attribute('word', 'p'))

        # load macros and wordlists
        self.lib_path = lib_path
        if self.lib_path:
            self.read_lib(self.lib_path)

    def describe_corpus(self):
        print("s-attributes without annotation:")
        print(self.corpus_s_att_without)
        print("s-attributes with annotation:")
        print(self.corpus_s_att_with)
        print("p-attributes:")
        print(self.corpus_p_att)

    def read_lib(self, path_lib):
        """reads macros and worldists"""

        # wordlists
        wordlists = glob(os.path.join(path_lib, "wordlists", "*"))
        for wordlist in wordlists:
            name = wordlist.split("/")[-1].split(".")[0]
            abs_path = os.path.abspath(wordlist)
            cqp_exec = 'define $%s < "%s";' % (
                name, abs_path
            )
            self.cqp.Exec(cqp_exec)

        # macros
        macros = glob(os.path.join(path_lib, "macros", "*"))
        for macro in macros:
            abs_path = os.path.abspath(macro)
            cqp_exec = 'define macro < "%s";' % abs_path
            self.cqp.Exec(cqp_exec)

    def _df_node_from_query(self, query, s_break, context=50,
                            match_strategy='standard'):
        """see df_node_from_query, which is a cached version of this method"""
        # matching strategy
        self.cqp.Exec('set MatchingStrategy "%s";' % match_strategy)

        # run query
        query += " within %s;" % s_break
        self.cqp.Query(query)
        cqp_result = self.cqp.Dump()

        # return empty dataframe if query has no results
        if cqp_result == [['']]:
            return DataFrame()

        # transform to dataframe
        df = DataFrame(cqp_result)
        df.columns = ['match_start', 'match_end', 'keyword', 'target']
        df = df.astype(int)

        # confine regions
        df_node = self.confine_df_node(df, s_break, context)

        return df_node

    def df_node_from_query(self, query, s_break, context=50,
                           match_strategy='standard'):
        """executes query within s_break to get df_node

        :param str query: valid CQP query (without 'within' clause)
        :param str s_break: s-attribute to confine regions
        :param int context: maximum context around node (symmetric)

        :return: df_node with 5 columns (match_start, match_end,
        region_start, region_end, s_id)
        :rtype: pd.DataFrame
        """

        # cache
        idx = create_identifier(self.corpus_name, self.lib_path,
                                query, s_break, context, match_strategy)
        cached_data = self.cache.get(idx)

        # create result
        if cached_data is None:
            df_node = self._df_node_from_query(query, s_break, context=50,
                                               match_strategy='standard')
            self.cache.set(idx, df_node)
        else:
            df_node = cached_data
        return df_node

    def _df_anchor_from_query(self, anchor_query, s_break, context=50,
                              match_strategy="standard"):
        """see df_anchor_from_query, which is a cached version of this method"""

        # check which anchors there are
        anchors = anchor_query_to_anchors(anchor_query)

        # matching strategy
        self.cqp.Exec('set MatchingStrategy "%s";' % match_strategy)

        # first run: 0 and 1
        self.cqp.Exec("set AnchorNumberTarget 0; set AnchorNumberKeyword 1;")
        self.cqp.Exec("Result = " + anchor_query + " within " + s_break + ";")

        dump = self.cqp.Exec("dump Result;")

        # convert to df_anchor
        df_anchor = dump_to_df_node(dump)
        df_anchor.columns = [0, 1]

        # if there's nothing to return ...
        if df_anchor.empty:
            return DataFrame()

        # join all the other ones
        for pair in [(2, 3), (4, 5), (6, 7), (8, 9)]:
            if pair[0] in anchors or pair[1] in anchors:
                # load matches
                self.cqp.Exec("Result;")
                # set appropriate anchors
                self.cqp.Exec(
                    "set AnchorNumberTarget %d; set AnchorNumberKeyword %d;" % pair
                )
                self.cqp.Exec('Temp = <match> ( %s );' % anchor_query)
                dump = self.cqp.Exec("dump Temp;")
                df = dump_to_df_node(dump)
                df.columns = [pair[0], pair[1]]
                df_anchor = df_anchor.join(df)

        # re-set CQP
        self.cqp.Exec(self.corpus_name)
        self.cqp.Exec("set ant 0; ank 1;")

        # NA handling
        df_anchor.dropna(axis=1, how='all', inplace=True)
        df_anchor = df_anchor.apply(to_numeric, downcast='integer')
        df_anchor.fillna(-1, inplace=True)

        # drop constant columns (contain only -1)
        df_anchor = df_anchor.loc[:, (df_anchor != df_anchor.iloc[0]).any()]

        # move index to columns
        df_anchor = df_anchor.reset_index()

        # confine regions
        df_anchor = self.confine_df_node(df_anchor, s_break, context)

        return df_anchor

    def df_anchor_from_query(self, anchor_query, s_break="text",
                             context=50, match_strategy="standard"):
        """executes anchored query within s_break to get df_anchor

        :param str anchor_query: anchored CQP query (without 'within' clause)
        :param str s_break: s-attribute to confine regions
        :param int context: maximum context around match (symmetric)

        :return: df_anchor with 5 base columns (match_start, match_end,
        region_start, region_end, s_id) and additional columns for each anchor
        :rtype: pd.DataFrame
        """

        # cache
        idx = create_identifier(self.corpus_name, self.lib_path,
                                anchor_query, s_break, context, match_strategy)
        cached_data = self.cache.get(idx)

        # create result
        if cached_data is None:
            df_anchor = self._df_anchor_from_query(anchor_query, s_break, context=50,
                                                   match_strategy='standard')
            self.cache.set(idx, df_anchor)
        else:
            df_anchor = cached_data

        return df_anchor

    def confine_df_node(self, df, s_break, context):

        # get s_break info
        s_regions = self.corpus.attribute(s_break, "s")
        s_region = DataFrame(
            df.match_start.apply(lambda x: s_regions.find_pos(x)).tolist()
        )
        df['s_start'] = s_region[0]
        df['s_end'] = s_region[1]
        df['s_id'] = df.match_start.apply(lambda x: s_regions.cpos2struc(x))

        # confine regions by s_break
        df['region_start'] = df.match_start - context
        df['region_end'] = df.match_end + context
        df.region_start = df[['region_start', 's_start']].max(axis=1)
        df.region_end = df[['region_end', 's_end']].min(axis=1)
        df = df.drop(['s_start', 's_end'], axis=1)

        # downcast values
        df = df.apply(to_numeric, downcast='integer')

        return df

    def count_matches(self, df_dump, p_att="lemma", split=True):
        """counts strings or tokens in df_dump[['match_start', 'match_end']]"""

        # undump the dump
        df_dump = df_dump.astype(str)
        dump = df_dump[['match_start', 'match_end']].values.tolist()
        self.cqp.Undump('Last', table=dump)

        # retrieve context strings
        context_strings = self.cqp.Exec(
            "tabulate Last match .. matchend %s;" % p_att
        ).split("\n")

        # optionally split strings into tokens
        if split:
            tokens = list()
            for t in context_strings:
                tokens += t.split(" ")
        else:
            tokens = context_strings

        # count
        counts = Counter(tokens)
        counts = DataFrame.from_dict(counts, orient='index')
        counts.columns = ['freq']
        counts = counts.sort_values('freq', ascending=False)
        return(counts)

    def lexicalize_positions(self, positions, p_att='word', ignore_missing=True):
        """Fills corpus positions. Raises IndexError if out-of-bounds.

        :param list positions: corpus positions to fill
        :param str p_att: p-attribute to fill positions with

        :return: lexicalizations of the positions
        :rtype: list

        """
        tokens_all = self.corpus.attribute(p_att, 'p')
        tokens_requested = list()
        if ignore_missing:
            positions = [p for p in positions if p != -1]
        for position in positions:
            tokens_requested.append(
                tokens_all[position]
            )
        return tokens_requested

    def get_marginals(self, items, p_att='word', regex=False, flags=1):
        """Extracts marginal frequencies for given items (0 if not in corpus).

        :param list items: items to get marginals for
        :param str p_att: p-attribute to get counts for
        :param int flags: 1 = %c, 2 = %d, 3 = %cd

        :return: counts for each item (indexed by items, column "f2")
        :rtype: DataFrame

        """
        tokens_all = self.corpus.attribute(p_att, 'p')
        counts = list()
        for item in items:
            if not regex:
                try:
                    counts.append(tokens_all.frequency(item))
                except KeyError:
                    counts.append(0)
            else:
                cpos = tokens_all.find_pattern(item, flags=flags)
                counts.append(len(cpos))
        f2 = DataFrame(index=items)
        f2['f2'] = counts
        return f2



import shelve
from hashlib import sha256


class Cache:
    def __init__(self, path="/tmp/ccc-cache"):
        self.path = path

    def get(self, key):
        with shelve.open(self.path) as db:
            try:
                return db[key]
            except KeyError:
                return None

    def set(self, key, value):
        with shelve.open(self.path) as db:
            db[key] = value


def create_identifier(*args):
    """
    Generate a hash from a string.

    :param args: arguments to use for identifier
    :return: hash for a given string
    :rtype: str
    """

    string = ' '.join([str(elem) for elem in args])
    return sha256(str(string).encode()).hexdigest()



import setuptools
from os import path


here = path.abspath(path.dirname(__file__))
with open(path.join(here, 'README.md')) as fh:
    long_description = fh.read()


setuptools.setup(
    name="cwb-ccc",
    version="0.9.2",
    author="Philipp Heinrich",
    author_email="philipp.heinrich@fau.de",
    description="CWB wrapper to extract concordances and collocates",
    packages=setuptools.find_packages(),
    long_description=long_description,
    long_description_content_type="text/markdown",
    url="https://gitlab.cs.fau.de/pheinrich/ccc",
    install_requires=[
        "pandas>=0.25.3",
        "cwb-python>=0.2.2"
    ],
    classifiers=[
        "Development Status :: 3 - Alpha",
        "Programming Language :: Python :: 3",
        "License :: OSI Approved :: GNU General Public License v3 or later (GPLv3+)",
        "Operating System :: OS Independent",
    ],
    python_requires='>=3.5',
)
